 (**********************************************************************)
 (*                                                                    *)
 (* Function:    ALARM - Special Alarm Clock Function                  *)
 (*                                                                    *)
 (* Creation Date: 2000-02-22                From: NEW                 *)
 (*                                                                    *)
 (* Author:  Eric Dnges                                               *)
 (*                                                                    *)
 (* Description:                                                       *)
 (* ------------                                                       *)
 (* This function provides a 32 bit alarm clock with a fixed reference *)
 (* time. This time can be reset any time; all alarm times are relative*)
 (* to this value.                                                     *)
 (*                                                                    *)
 (* Parameters:                                                        *)
 (*                                                                    *)
 (* PRM0: upper 16 bits of currently programmed alarm time             *)
 (* PRM1: lower 16 bits of currently programmed alarm time             *)
 (* PRM2: upper 16 bits of current time                                *)
 (* PRM3: lower 16 bits of current time                                *)
 (* PRM4: TCR1 base time everything else is based on                   *)
 (*                                                                    *)
 (* The CPU should only write to PRM2 before initialization. PRM3      *)
 (* should not be written at all. PRM0 and PRM1 should not be changed  *)
 (* until the alarm has been canceled or has expired.                  *)
 (*                                                                    *)
 (* Updates:   By:   Modification:                                     *)
 (* --------   ---   -------------                                     *)
 (*                                                                    *)
 (*--------------------------------------------------------------------*)
 (* Standard Exits Used:-   End_Of_Phase: Y         End_Of_Link: N     *)
 (*                                                                    *)
 (* External Files %included: NONE.                                    *)
 (*                                                                    *)
 (* CODE SIZE excluding standard exits = ?? LONG WORDS                 *)
 (*--------------------------------------------------------------------*)
 (*                                                                    *)
 (* This Revision:  1.1                                                *)
 (* LAST MODIFIED: 2000-03-31      BY: Eric Dnges                     *)
 (*                                                                    *)
 (* This code is in the public domain, i.e. you may use it in any way  *)
 (* you see fit. However, it comes with absolutely no warranties. Use  *)
 (* at your own risk !                                                 *)
 (**********************************************************************)
 
 (* Parameters *)
 %macro ALARM_WHEN_HI	'prm0'. (* upper 16 bits of next alarm time   *)
 %macro ALARM_WHEN_LO	'prm1'. (* lower 16 bits of next alarm time   *)
 %macro ALARM_TIME_HI	'prm2'. (* upper 16 bits of last matched time *)
 %macro ALARM_TIME_LO	'prm3'. (* lower 16 bits of last matched time *)
 %macro ALARM_BASE	'prm4'.	(* TCR1 base time                     *)
 %macro TFLAG            'flag0'. (* set == dummy event                *)
 %macro VFLAG            'flag1'. (* set == valid time in WHEN         *)
 
 
 (* Subroutines *)
 
 
 (**********************************************************************)
 (*                                                                    *)
 (* ALARM_NextMatch                                                    *)
 (*                                                                    *)
 (* Calculate the next match time. Use the following algorithm:        *)
 (*                                                                    *)
 (* if (VFLAG) {                                                       *)
 (*   if (when < time) {                                               *)
 (*     TFLAG := 0                                                     *)
 (*     ert := time_lo + base (immediate match)                        *)
 (*   }                                                                *)
 (*   else if (when - time > $8000)                                    *)
 (*     TFLAG := 1                                                     *)
 (*     ert := time_lo + $8000 + base                                  *)
 (*   }                                                                *)
 (*   else {                                                           *)
 (*     TFLAG := 0                                                     *)
 (*     ert := when_lo + base                                          *)
 (*   }                                                                *)
 (* }                                                                  *)
 (* else {                                                             *)
 (*   TFLAG := 1                                                       *)
 (*   ert := time_lo + $8000 + base                                    *)
 (* }                                                                  *)
 (*                                                                    *)
 (*                                                                    *)
 (**********************************************************************)
 
 ALARM_NextMatch:
 	(* First, check if we have a new valid alarm time. If not,    *)
 	(* we always create a dummy event.                            *)
 
 	if @VFLAG = FALSE then goto ALARM_NextDummy.
 
 	(* Do a 32 bit subtraction to check if when <= time. Note how *)
 	(* we use the branch delay slot to load p, which is needed in *)
 	(* ALARM_NextDummy as well.                                   *)
 
 	ram p <- @ALARM_TIME_LO.
 
 	(* load WHEN_LO and WHEN_HI as a 32 bit coherent access       *)
 	ram diob <- @ALARM_WHEN_HI.
 	au a := diob; ram diob <- @ALARM_WHEN_LO.
 
 	(* Now, do the actual subtraction (lower 16 bits)             *)
 	au sr := diob - p, ccl;	ram p <- @ALARM_TIME_HI.
 
 	if C = TRUE then goto ALARM_NextXX, flush.
 
 	(* Again, make use of a branch delay slot                     *)
 	goto ALARM_NextYY.
 	au a := a - p, ccl.
 
 
 ALARM_NextXX:
 	au a := a - p - 1, ccl.
 
 
 ALARM_NextYY:
 	(* If when < time, the next match should occur immediatly     *)
 	(* This should be '<=', but that case is much harder to handle*)
 	(* for 32 bits, and not really necessary (since the match will*)
 	(* be scheduled immediately anyway)                           *)
 
 	if C = TRUE then goto ALARM_NextImm, flush.
 
 	(* Now, check if the difference between when and time is      *)
 	(* smaller than $8000. If it is, we do not need to insert a   *)
 	(* dummy event. Note how we have to use an addition because   *)
 	(* the TPU does not support a subtraction for this mode.      *)
 
 	au sr := sr + $8000, ccl.
 
 	if C = TRUE then goto ALARM_NextDummy.
 	ram p <- @ALARM_TIME_LO.         (* use the branch delay slot *)
 
 	(* Since the lower 16 bits were smaller than $8000, we need   *)
 	(* to check if the upper 16 bits are zero                     *)
 
 	au a := a, ccl.
 	if Z = TRUE then goto ALARM_NextEvent, flush.
 
 
 ALARM_NextDummy:
 	(* p was already loaded with ALARM_TIME_LO *)
 	(* The strange instruction order is to maximize the use of    *)
 	(* branch delay slots.                                        *)
 	au ert := p + $8000; ram p <- @ALARM_BASE.
 	return.
 	au ert := ert + p; chan set @TFLAG.
 
 
 ALARM_NextEvent:
 	ram p <- @ALARM_WHEN_LO.
 	ram diob <- @ALARM_BASE; return.
 	au ert := p + diob; chan clear @TFLAG.
 
 
 ALARM_NextImm:
 	(* Cause an immediate match event *)
 	return.
 	au ert := tcr1; chan clear @TFLAG.
 
 
 (**********************************************************************)
 (*                                                                    *)
 (* ALARM_SetTime sets time_hi and time_lo to the current values. It   *)
 (* assumes that p is preloaded with ALARM_BASE on entry and ert with  *)
 (* the current value of tcr1 (or with the current match time)         *)
 (*                                                                    *)
 (**********************************************************************)
 
 ALARM_SetTime:
 	au diob := ert - p; ram p <- @ALARM_TIME_LO.
 
 	(* check if the new TIME_LO is smaller than the old TIME_LO.  *)
 	(* If it is, then we need to carry into the upper 16 bits     *)
 	(* The instruction order is a bit strange to try to get a     *)
 	(* coherent 32 bit write for the time value. Note how the     *)
 	(* branch delay slot is used to this end.                     *)
 
 	au nil := diob - p, ccl. ; ram p <- @ALARM_TIME_HI.
 
 	if C = FALSE then goto ALARM_SetTimeEnd.
 	ram diob -> @ALARM_TIME_LO.
 
 	au p := p + 1; ram p -> @ALARM_TIME_HI.
 
 ALARM_SetTimeEnd:
 	return, flush.
 
 
 (**********************************************************************)
 (*                                                                    *)
 (* STATE :   ALARM_Init                                               *)
 (*                                                                    *)
 (* PRELOAD PARAMETER :                                                *)
 (*                                                                    *)
 (* ENTER WHEN : HSR = 01                                              *)
 (*                                                                    *)
 (* ACTION : Initialize the alarm timer. Depending on the HSQ bits, do *)
 (*	    the following:                                            *)
 (*                                                                    *)
 (*          HSQ1 HSQ0                                                 *)
 (*             0    0  - update time_lo and time_hi                   *)
 (*             1    0  - reset the time base                          *)
 (*             1    1  - reset the alarm, but use a time base from a  *)
 (*                       different channel for the base.              *)
 (*                                                                    *)
 (**********************************************************************)
 %entry start_address *; disable_match;
 name = ALARM_Init;
 cond hsr1 = 0, hsr0 = 1.
 
 ALARM_Init :
 	(* Setup the channel pin as a permanent high output *)
 	chan pac := no_change, PIN := high, tbs := out_m1_c1, enable_mtsr.
 
 	if HSQ1 = FALSE then goto ALARM_GetTime, flush.
 
 	if HSQ0 = FALSE then goto ALARM_WriteBase.
 
 	(* Use the branch delay slot. Only one of the values is used, *)
 	(* depending on if the branch is taken or not                 *)
 	au p := tcr1; ram diob <- @ALARM_BASE.
 
 	(* Interpret ALARM_BASE as a pointer to a reference value of  *)
 	(* another channel                                            *)
 	ram p <- by_diob.
 
 
 ALARM_WriteBase:
 	(* write the alarm base and clear the VFLAG *)
 	ram p -> @ALARM_BASE; chan clear @VFLAG.
 	au p := 0; ram p -> @ALARM_TIME_LO.
 	ram p -> @ALARM_TIME_HI.
 
 	(* Fall into GetTime                                          *)
 
 
 ALARM_GetTime:
 	(* Again, note the use of the branch delay slot               *)
 	call ALARM_SetTime.
 	au ert := tcr1; ram p <- @ALARM_BASE.
 
 	(* Now, recalculate the next match event if there is no valid *)
 	(* alarm time programmed. This is to prevent unnecessary dummy*)
 	(* events                                                     *)
 	if @VFLAG = TRUE then goto ALARM_GetTimeEnd, flush.
 
 	call ALARM_NextMatch, flush.
 
 	chan write_MER, neg_MRL, neg_TDL, neg_LSL; end.
 
 ALARM_GetTimeEnd:
 	end.
 
 
 (**********************************************************************)
 (*                                                                    *)
 (* STATE :   ALARM_SetAlarm                                           *)
 (*                                                                    *)
 (* PRELOAD PARAMETER :                                                *)
 (*                                                                    *)
 (* ENTER WHEN : HSR = 10                                              *)
 (*                                                                    *)
 (* ACTION : Set a new alarm time                                      *)
 (*                                                                    *)
 (**********************************************************************)
 %entry name = ALARM_SetAlarm; start_address *; disable_match;
 cond hsr1 = 1, hsr0 = 0.
 
 ALARM_SetAlarm:
 	(* Set the current time. Note how we maximize the use of      *)
 	(* delay slots                                                *)
 	call ALARM_SetTime.
 	au ert := tcr1; ram p <- @ALARM_BASE.
 
 	chan set @VFLAG; call ALARM_NextMatch, flush.
 
 	(* For timing: force pin high again *)
 	chan PIN := high, write_MER; end.
 
 
 (**********************************************************************)
 (*                                                                    *)
 (* STATE :   ALARM_ClearAlarm                                         *)
 (*                                                                    *)
 (* PRELOAD PARAMETER :                                                *)
 (*                                                                    *)
 (* ENTER WHEN : HSR = 11                                              *)
 (*                                                                    *)
 (* ACTION : Cancel the current alarm request (and schedule a dummy    *)
 (*          event.                                                    *)
 (*                                                                    *)
 (**********************************************************************)
 %entry name = ALARM_ClearAlarm; start_address *; disable_match;
 cond hsr1 = 1, hsr0 = 1.
 
 
 ALARM_ClearAlarm:
 	call ALARM_SetTime.
 	au ert := tcr1; ram p <- @ALARM_BASE.
 
 	chan clear @VFLAG; call ALARM_NextMatch, flush.
 
 	(* For timing: force pin high again *)
 	chan PIN := high, write_MER; end.
 
 
 (**********************************************************************)
 (*                                                                    *)
 (* STATE :   ALARM_MatchEvent                                         *)
 (*                                                                    *)
 (* PRELOAD PARAMETER :                                                *)
 (*                                                                    *)
 (* ENTER WHEN : HSR = 00 M/TSR = 1 TFLAG = 0                          *)
 (*                                                                    *)
 (* ACTION : Do a match for a real event.                              *)
 (*                                                                    *)
 (**********************************************************************)
 %entry name = ALARM_MatchEvent; start_address *; disable_match;
 cond hsr1 = 0, hsr0 = 0, m/tsr = 1, @TFLAG = 0.
 
 ALARM_MatchEvent:
 	ram p <- @ALARM_BASE; call ALARM_SetTime, flush.
 
 	chan clear @VFLAG; call ALARM_NextMatch, flush.
 
 	(* for timing purposes: force pin low on ALARM *)
 	chan PIN := low, write_MER, neg_MRL, neg_TDL, neg_LSL, cir; end.
 
 
 (**********************************************************************)
 (*                                                                    *)
 (* STATE :   ALARM_MatchDummy                                         *)
 (*                                                                    *)
 (* PRELOAD PARAMETER :                                                *)
 (*                                                                    *)
 (* ENTER WHEN : HSR = 00 M/TSR = 1 TFLAG = 1                          *)
 (*                                                                    *)
 (* ACTION : Do a match for a dummy event.                             *)
 (*                                                                    *)
 (**********************************************************************)
 %entry name = ALARM_MatchDummy; start_address *; disable_match;
 cond hsr1 = 0, hsr0 = 0, m/tsr = 1, @TFLAG = 1.
 
 ALARM_MatchDummy:
 	(* Update the current time *)
 	ram p <- @ALARM_BASE; call ALARM_SetTime, flush.
 
 	(* Calculate the next match time *)
 	call ALARM_NextMatch, flush.
 
 	(* Set up the next match *)
 	chan write_MER, neg_MRL; end.
 
 
 (**********************************************************************)
 (*     UNUSED ENTRIES - execute an end.                               *)
 (**********************************************************************)
 %entry name = ALARM_Unused; start_address End_of_Link;
 cond hsr1 = 0, hsr0 = 0, lsr = 1, m/tsr = 0.
 
